/*
 * Copyright (c) 2009,2010,2018 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <linux/joystick.h>

#include <rmp/rmpLib.h>

#define  SIGN(x)             ((x)<0. ? -1. : 1.)
#define  C_DEG_RAD           (M_PI/180.)
#define  DEG_TO_RAD(x)       ((x) * C_DEG_RAD)
#define  C_RAD_DEG           (180./M_PI)
#define  RAD_TO_DEG(x)       ((x) * C_RAD_DEG)

/* For joystick driver */
#define NAME_LENGTH 128

unsigned char axes = 2;		/* number of joystick axes */
unsigned char buttons = 2;	/* number of joystick buttons */
int *axis;			/* array of joystick axis values */
int *button;			/* array of joystick button state */
short v, w;			/* linear & angular speed references */
int quit;			/* quit flag */

int version = 0x000800;
char name[NAME_LENGTH] = "Unknown";

pthread_mutex_t joymutex = PTHREAD_MUTEX_INITIALIZER;

/* From SI to RMP values */
#define RMP400_CONVERT_TO_RMP_SPEED(x)   (-(x)*332.0*1.29)
#define RMP400_CONVERT_TO_RMP_ANGLE(x)   (RAD_TO_DEG((x)*17.16))

/* Speed gains for joystick control */
#define VGAIN	1./(32768.0*32768.0) /* linear  */
#define WGAIN   1./(32768.0*32768.0) /* angular */


/*----------------------------------------------------------------------*/
/**
 ** Init joystick
 **/
static int
joyinit(char *dev)
{
	int joy;

	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return -1;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);

	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16,
	       (version >> 8) & 0xff, version & 0xff);
	return joy;
}
/*----------------------------------------------------------------------*/

/**
 ** task reading data from the joystick and producing references;
 **/
void *
joyRead(void *param)
{
	int fd = (int)(intptr_t)param;
	struct js_event js;
	struct pollfd pfd;
	int ret;

	while (!quit) {
		pfd.fd = fd;
		pfd.events = POLLIN;
		pfd.revents = 0;
		ret = poll(&pfd, 1, 100);

		if (ret < 0) {
			if (errno != EINTR)
				fprintf(stderr, "poll error: %s\n",
					strerror(errno));
			continue;
		}
		if (ret == 0)
			/* nothing to read from joystick */
			continue;
		if ((pfd.revents & POLLIN) != POLLIN)
			continue;
		/* something to read from joystick */
		if (read(fd, &js, sizeof(struct js_event))
		    != sizeof(struct js_event)) {
			perror("\njstest: error reading");
			v = 0;
			w = 0;
			break;
		}
		switch(js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			button[js.number] = js.value;
			printf("button %d %d\n", js.number, js.value);
			break;
		case JS_EVENT_AXIS:
			axis[js.number] = js.value;
			printf("axis %d %d\n", js.number, js.value);
			break;
		}
		pthread_mutex_lock(&joymutex);
		/* Test for motion button */
		if (button[0]) {
			/* compute speed refs */
			v = -SIGN(axis[1]) * 
			    RMP400_CONVERT_TO_RMP_SPEED(axis[1]*axis[1]*VGAIN);
			w = -SIGN(axis[0]) *
			    RMP400_CONVERT_TO_RMP_ANGLE(axis[0]*axis[0]*WGAIN);
			printf("\t\tmove %d %d\n", v, w);
		} else {
			/* Set refs to zero */
			printf("stop\n");
			v = 0;
			w = 0;
		}
		pthread_mutex_unlock(&joymutex);
	}
	return NULL;
}
/*----------------------------------------------------------------------*/
/**
 ** Main loop of joystick control mode
 **/
static void
joymove(struct RMP_DEV_STR **rmp)
{
	int i;
	int retval;
	struct timespec ts;

	while (!quit) {
		pthread_mutex_lock(&joymutex);
		for (i = 0; i < 2; i++) {
			if (rmp[i] != NULL)
				retval = rmpSendSpeed(rmp[i], v, w);
			if (retval != 0)
				break;
		}
		pthread_mutex_unlock(&joymutex);
		if (retval != 0) {
			fprintf(stderr, "Are the motors on?\n");
			fprintf(stderr, "aborting.\n");
			quit = 1;
		}
		/* Test for quit button */
		if (button[3])
			quit = 1;
		ts.tv_sec = 0;
		ts.tv_nsec = 100000000; /* 100ms */
		nanosleep(&ts, NULL);
	} /* while */
} /* joymove */

/*----------------------------------------------------------------------*/

/**
 ** Task reading data from one RMP device.
 ** Frames are decoded and the device state is updated, but not used.
 **/
void *
rmpReadTask(void *ptr)
{
	struct RMP_DEV_STR *dev = (struct RMP_DEV_STR *)ptr;
	struct timespec ts;

	while (!quit) {
		if (rmpReadPackets(dev) < 0)
			break;
		ts.tv_sec = 0;
		ts.tv_nsec = 100000000; /* 100ms */
		nanosleep(&ts, NULL);
	}
	return NULL;
}

/*----------------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
	struct RMP_DEV_STR **rmp;
	pthread_t rid[2], jid;
	int i, j;

	v = 0;
	w = 0;
	quit = 0;

	/* Connect to RMP devices */
	if (rmpOpenAll(&rmp) < 0)
		exit(2);
	printf("rmpOpen done\n");

	/* spawn RMP reader tasks */
	for (i = 0; i < 2; i++) {
		pthread_create(&rid[i], NULL, rmpReadTask, (void *)rmp[i]);
		pthread_detach(rid[i]);
	}

	/* Connect to joystick */
	j = joyinit("/dev/input/js0");
	if (j < 0) {
		fprintf(stderr, "Error initializing joystick\n");
		exit(2);
	}
	printf("joyinit done\n");
	axis = calloc(axes, sizeof(int));
	button = calloc(buttons, sizeof(int));

	printf("use button 4 to quit\n"
	       "   button 1 + left stick to move\n");

	pthread_create(&jid, NULL, joyRead, (void *)(intptr_t)j);
	pthread_detach(jid);
	printf("go\n");

	/* launch main loop */
	joymove(rmp);

	/* Free resources */
	free(button);
	free(axis);
	close(j);
	for (i = 0; i < 2; i++) {
		if (rmp[i] != NULL)
			rmpClose(rmp[i]);
	}
	return 0;
}
